import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import re
from collections import defaultdict
from sklearn.metrics import cohen_kappa_score
from typing import List

"""
This file calculates cohen kappa and brennan prediger scores for the coding
It prints all values into the terminal


@author: Jonathan van Oudheusden
@date: 2024-06-19

Required files: 
    - Output/Intro_codes.csv
    - Output/Intro_peer_codes.csv
    Both need: 'rand_id'
    Intro_codes needs: 'Codes'
    Intro_peer_codes needs: 'Peer_codes'
    
Output files:
- It prints all values into the terminal
The output is also in CalculateCodingAgreementResults.txt
"""

# Load both coding datasets
coded = pd.read_csv('Output/Intro_codes.csv')  
peer_coded = pd.read_csv('Output/Intro_peer_codes.csv')  

# Merge datasets on 'rand_id' column
dfAllCodes = pd.merge(coded, peer_coded[['rand_id', 'Peer_codes']], on='rand_id', how='inner')

# Function to split codes by comma and remove any surrounding whitespace
def splitCodes(Codes):
    return [code.strip() for code in Codes.split(',')]

# Function to calculate Brennan-Prediger agreement score
def brennan_prediger(ratings1, ratings2, num_categories):
    if len(ratings1) != len(ratings2):
        raise ValueError("The number of ratings from both raters must be the same.")
    ratings1 = np.array(ratings1)
    ratings2 = np.array(ratings2)
    total_items = len(ratings1)
    agreements = np.sum(ratings1 == ratings2)
    A_o = agreements / total_items
    A_e = 1 / num_categories
    BP = 2 * A_o - 1
    return BP

# Create a dictionary to store counts and scores of codes
allCodesDict = defaultdict(lambda: [[0, 0], 99, 99])

# Count occurrences of each code in both 'Codes' and 'Peer_codes'
for index, row in dfAllCodes.iterrows():
    mycodes = row['Codes']
    peerCodes = row['Peer_codes']

    # Check if 'mycodes' is not null and not empty
    if not pd.isnull(mycodes) and mycodes.strip() != '':
        for code in splitCodes(mycodes):
            if code != '':
                allCodesDict[code][0][0] += 1 # Increment count for main codes

    # Check if 'peerCodes' is not null and not empty
    if not pd.isnull(peerCodes) and peerCodes.strip() != '':
        for peerCode in splitCodes(peerCodes):
            if peerCode != '':
                allCodesDict[peerCode][0][1] += 1 # Increment count for peer codes

# List of all unique codes
allCodesList: List[str] = [code for code in allCodesDict.keys() if code != '']

print(f"Total number of unique codes: {len(allCodesList)}")

dfAllCodesFiltered = dfAllCodes.copy()

# Initialize columns for each code in the dataframe
for code in allCodesList:
    dfAllCodesFiltered[code] = 0
    dfAllCodesFiltered["Peer " + code] = 0

# Populate the dataframe with 1 if the code is present, otherwise 0
dfAllCodesFilteredcopy = dfAllCodesFiltered.copy()
for index, row in dfAllCodesFilteredcopy.iterrows():
    mycodes = row['Codes']
    peerCodes = row['Peer_codes']

    # Check if 'mycodes' is not null and not empty
    if not pd.isnull(mycodes) and mycodes.strip() != '':
        for code in splitCodes(mycodes):
            if code in dfAllCodesFilteredcopy.columns:
                dfAllCodesFilteredcopy.at[index, code] = 1 # Mark presence of main code

    # Check if 'peerCodes' is not null and not empty
    if not pd.isnull(peerCodes) and peerCodes.strip() != '':
        for peerCode in splitCodes(peerCodes):
            if "Peer " + peerCode in dfAllCodesFilteredcopy.columns:
                dfAllCodesFilteredcopy.at[index, 'Peer ' + peerCode] = 1 # Mark presence of peer code

# Variables to store total weights and weighted sums of kappa and Brennan-Prediger scores
total_weight = 0
weighted_kappa_sum = 0
weighted_bp_sum = 0

# calculated cohen's kappa and brennan prediger for each individual code and weighed average
for code in allCodesList:
    kappa = cohen_kappa_score(dfAllCodesFilteredcopy[code], dfAllCodesFilteredcopy['Peer ' + code])
    bp = brennan_prediger(dfAllCodesFilteredcopy[code], dfAllCodesFilteredcopy['Peer ' + code], 2)
    allCodesDict[code][1] = kappa
    allCodesDict[code][2] = bp
    weight = sum(allCodesDict[code][0])
    total_weight += weight
    weighted_kappa_sum += kappa * weight
    weighted_bp_sum += bp * weight

# Lists to store kappa and Brennan-Prediger scores for each code
Kappalist = []
BPList = []

# print out scores and values of individual codes, sorted from most used to least
sorted_dict = dict(sorted(allCodesDict.items(), key=lambda item: sum(item[1][0]), reverse=True))
for code, count in sorted_dict.items():
    print(f"{code}, Count: {count[0]} kappa: {count[1]} bp: {count[2]}")
    Kappalist.append(count[1])
    BPList.append(count[2])

# Calculate weighted average kappa and Brennan-Prediger scores
weighted_avg_kappa = weighted_kappa_sum / total_weight
weighted_avg_bp = weighted_bp_sum / total_weight

print("Weighted Average Cohen Kappa:")
print(weighted_avg_kappa)

print("Weighted Average Brennan-Prediger:")
print(weighted_avg_bp)